"""Exceptions raised by the Route53 service."""

from typing import Any, Optional

from moto.core.exceptions import RESTError


class Route53ClientError(RESTError):
    """Base class for Route53 errors."""

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault("template", "wrapped_single_error")
        super().__init__(*args, **kwargs)


class ConflictingDomainExists(Route53ClientError):
    """Domain already exists."""

    code = 400

    def __init__(self, domain_name: str, delegation_set_id: Optional[str]) -> None:
        message = (
            f"Cannot create hosted zone with DelegationSetId DelegationSetId:{delegation_set_id} as the DNSName"
            f"{domain_name} conflicts with existing ones sharing the delegation set"
        )
        super().__init__("ConflictingDomainExists", message)


class InvalidInput(Route53ClientError):
    """Malformed ARN for the CloudWatch log group."""

    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidInput", message)


class UnsupportedCharacter(Route53ClientError):
    """Malformed ARN for the CloudWatch log group."""

    code = 400

    def __init__(self, code: str, char: str):
        super().__init__(
            code,
            f"FATAL problem: UnsupportedCharacter (Value contains unsupported characters) encountered with '{char}'",
        )


class InvalidCloudWatchArn(InvalidInput):
    def __init__(self) -> None:
        message = "The ARN for the CloudWatch Logs log group is invalid"
        super().__init__(message)


class InvalidActionValue(InvalidInput):
    def __init__(self, value: str):
        message = (
            f"Invalid XML ; cvc-enumeration-valid: Value '{value}' is not facet-valid"
            " with respect to enumeration '[CREATE, DELETE, UPSERT]'. It must be a value from the enumeration."
        )
        super().__init__(message)


class InvalidPaginationToken(Route53ClientError):
    """Bad NextToken specified when listing query logging configs."""

    code = 400

    def __init__(self) -> None:
        message = (
            "Route 53 can't get the next page of query logging configurations "
            "because the specified value for NextToken is invalid."
        )
        super().__init__("InvalidPaginationToken", message)


class InvalidVPCId(Route53ClientError):
    """Missing/Invalid VPC ID"""

    code = 400

    def __init__(self) -> None:
        message = "Invalid or missing VPC Id."
        super().__init__("InvalidVPCId", message)
        self.content_type = "text/xml"


class NoSuchCloudWatchLogsLogGroup(Route53ClientError):
    """CloudWatch LogGroup has a permissions policy, but does not exist."""

    code = 404

    def __init__(self) -> None:
        message = "The specified CloudWatch Logs log group doesn't exist."
        super().__init__("NoSuchCloudWatchLogsLogGroup", message)


class NoSuchHostedZone(Route53ClientError):
    """HostedZone does not exist."""

    code = 404

    def __init__(self, host_zone_id: str):
        message = f"No hosted zone found with ID: {host_zone_id}"
        super().__init__("NoSuchHostedZone", message)
        self.content_type = "text/xml"


class NoSuchHealthCheck(Route53ClientError):
    """HealthCheck does not exist."""

    code = 404

    def __init__(self, health_check_id: str):
        message = f"A health check with id {health_check_id} does not exist."
        super().__init__("NoSuchHealthCheck", message)
        self.content_type = "text/xml"


class HostedZoneNotEmpty(Route53ClientError):
    """HostedZone does not exist."""

    code = 400

    def __init__(self) -> None:
        message = (
            "The hosted zone contains resource records that are not SOA or NS records."
        )
        super().__init__("HostedZoneNotEmpty", message)
        self.content_type = "text/xml"


class PublicZoneVPCAssociation(Route53ClientError):
    """Public hosted zone can't be associated."""

    code = 400

    def __init__(self) -> None:
        message = "You're trying to associate a VPC with a public hosted zone. Amazon Route 53 doesn't support associating a VPC with a public hosted zone."
        super().__init__("PublicZoneVPCAssociation", message)
        self.content_type = "text/xml"


class LastVPCAssociation(Route53ClientError):
    """Last VPC can't be disassociate."""

    code = 400

    def __init__(self) -> None:
        message = "The VPC that you're trying to disassociate from the private hosted zone is the last VPC that is associated with the hosted zone. Amazon Route 53 doesn't support disassociating the last VPC from a hosted zone."
        super().__init__("LastVPCAssociation", message)
        self.content_type = "text/xml"


class NoSuchQueryLoggingConfig(Route53ClientError):
    """Query log config does not exist."""

    code = 404

    def __init__(self) -> None:
        message = "The query logging configuration does not exist"
        super().__init__("NoSuchQueryLoggingConfig", message)


class QueryLoggingConfigAlreadyExists(Route53ClientError):
    """Query log config exists for the hosted zone."""

    code = 409

    def __init__(self) -> None:
        message = "A query logging configuration already exists for this hosted zone"
        super().__init__("QueryLoggingConfigAlreadyExists", message)


class InvalidChangeBatch(Route53ClientError):
    code = 400

    def __init__(self) -> None:
        message = "Number of records limit of 1000 exceeded."
        super().__init__("InvalidChangeBatch", message)


class NoSuchDelegationSet(Route53ClientError):
    code = 400

    def __init__(self, delegation_set_id: str):
        super().__init__("NoSuchDelegationSet", delegation_set_id)
        self.content_type = "text/xml"


class DnsNameInvalidForZone(Route53ClientError):
    code = 400

    def __init__(self, name: str, zone_name: str):
        error_msg = (
            f"""RRSet with DNS name {name} is not permitted in zone {zone_name}"""
        )
        super().__init__("InvalidChangeBatch", error_msg)


class ResourceRecordAlreadyExists(Route53ClientError):
    code = 400

    def __init__(self, name: str, _type: str):
        super().__init__(
            "InvalidChangeBatch",
            f"Tried to create resource record set [name='{name}', type='{_type}'] but it already exists",
        )



import unittest

import boto3

from moto import mock_aws
from tests import EXAMPLE_AMI_ID


class TestNestedDecoratorsBoto3(unittest.TestCase):
    @mock_aws
    def setup_sqs_queue(self) -> None:
        conn = boto3.resource("sqs", region_name="us-east-1")
        queue = conn.create_queue(QueueName="some-queue")

        queue.send_message(MessageBody="test message 1")

        queue.reload()
        assert queue.attributes["ApproximateNumberOfMessages"] == "1"

    @mock_aws
    def test_nested(self) -> None:
        self.setup_sqs_queue()

        conn = boto3.client("ec2", region_name="us-west-2")
        conn.run_instances(ImageId=EXAMPLE_AMI_ID, MinCount=1, MaxCount=1)


def test_multiple_mocks() -> None:
    with mock_aws():
        client = boto3.client("sqs", region_name="us-east-1")
        client.create_queue(QueueName="some-queue")

        # Starting another (inner) mock does not reset the data
        with mock_aws():
            assert len(client.list_queues()["QueueUrls"]) == 1

        # Ending an inner mock does not reset the data
        assert len(client.list_queues()["QueueUrls"]) == 1



import moto.server as server
from moto import mock_aws

"""
Test the different server responses
"""


@mock_aws
def test_kinesisvideo_server_is_up():
    backend = server.create_backend_app("kinesisvideo")
    test_client = backend.test_client()
    res = test_client.post("/listStreams")
    assert res.status_code == 200



import string
from typing import Any, Dict, List, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random as random
from moto.utilities.utils import get_partition


class Pipeline(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        name: str,
        input_bucket: str,
        output_bucket: str,
        role: str,
        content_config: Dict[str, Any],
        thumbnail_config: Dict[str, Any],
    ):
        a = "".join(random.choice(string.digits) for _ in range(13))
        b = "".join(random.choice(string.ascii_lowercase) for _ in range(6))
        self.id = f"{a}-{b}"
        self.name = name
        self.arn = f"arn:{get_partition(region)}:elastictranscoder:{region}:{account_id}:pipeline/{self.id}"
        self.status = "Active"
        self.input_bucket = input_bucket
        self.output_bucket = output_bucket or content_config["Bucket"]
        self.role = role
        self.content_config = content_config or {"Bucket": self.output_bucket}
        if "Permissions" not in self.content_config:
            self.content_config["Permissions"] = []
        self.thumbnail_config = thumbnail_config or {"Bucket": self.output_bucket}
        if "Permissions" not in self.thumbnail_config:
            self.thumbnail_config["Permissions"] = []

    def update(self, name: str, input_bucket: str, role: str) -> None:
        if name:
            self.name = name
        if input_bucket:
            self.input_bucket = input_bucket
        if role:
            self.role = role

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Id": self.id,
            "Name": self.name,
            "Arn": self.arn,
            "Status": self.status,
            "InputBucket": self.input_bucket,
            "OutputBucket": self.output_bucket,
            "Role": self.role,
            "Notifications": {
                "Progressing": "",
                "Completed": "",
                "Warning": "",
                "Error": "",
            },
            "ContentConfig": self.content_config,
            "ThumbnailConfig": self.thumbnail_config,
        }


class ElasticTranscoderBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.pipelines: Dict[str, Pipeline] = {}

    def create_pipeline(
        self,
        name: str,
        input_bucket: str,
        output_bucket: str,
        role: str,
        content_config: Dict[str, Any],
        thumbnail_config: Dict[str, Any],
    ) -> Tuple[Pipeline, List[str]]:
        """
        The following parameters are not yet implemented:
        AWSKMSKeyArn, Notifications
        """
        pipeline = Pipeline(
            self.account_id,
            self.region_name,
            name,
            input_bucket,
            output_bucket,
            role,
            content_config,
            thumbnail_config,
        )
        self.pipelines[pipeline.id] = pipeline
        warnings: List[str] = []
        return pipeline, warnings

    def list_pipelines(self) -> List[Dict[str, Any]]:
        return [p.to_dict() for _, p in self.pipelines.items()]

    def read_pipeline(self, pipeline_id: str) -> Pipeline:
        return self.pipelines[pipeline_id]

    def update_pipeline(
        self, pipeline_id: str, name: str, input_bucket: str, role: str
    ) -> Tuple[Pipeline, List[str]]:
        """
        The following parameters are not yet implemented:
        AWSKMSKeyArn, Notifications, ContentConfig, ThumbnailConfig
        """
        pipeline = self.read_pipeline(pipeline_id)
        pipeline.update(name, input_bucket, role)
        warnings: List[str] = []
        return pipeline, warnings

    def delete_pipeline(self, pipeline_id: str) -> None:
        self.pipelines.pop(pipeline_id)


elastictranscoder_backends = BackendDict(ElasticTranscoderBackend, "elastictranscoder")



import base64
import json
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Event, Thread
from typing import Optional, Tuple
from unittest import SkipTest
from uuid import uuid4

import boto3
import requests
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.x509 import load_pem_x509_certificate

from moto import mock_aws, settings

# Original Request
# https://github.com/getmoto/moto/discussions/7985
# AWS docs
# https://docs.aws.amazon.com/sns/latest/dg/sns-verify-signature-of-message.html
# Request for Boto3 support for this feature
# https://github.com/boto/boto3/issues/2508


class WebRequestHandler(BaseHTTPRequestHandler):
    @property
    def post_data(self):
        content_length = int(self.headers.get("Content-Length", 0))
        return self.rfile.read(content_length)

    def do_GET(self):
        msg = json.loads(self.post_data.decode("UTF-8"))
        topic_arn = msg["TopicArn"]
        TestHTTPMessageVerification.MESSAGES_RECEIVED[topic_arn] = msg
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(b"")

    def do_POST(self):
        self.do_GET()


class ThreadedSnsReceiver:
    def __init__(self):
        self._port = 0

        self._thread: Optional[Thread] = None
        self._ip_address = "0.0.0.0"
        self._server: Optional[HTTPServer] = None
        self._server_ready_event = Event()

    def _server_entry(self) -> None:
        self._server = HTTPServer(("0.0.0.0", 0), WebRequestHandler)
        self._server_ready_event.set()
        self._server.serve_forever()

    def start(self) -> None:
        self._thread = Thread(target=self._server_entry, daemon=True)
        self._thread.start()
        self._server_ready_event.wait()

    def get_host_and_port(self) -> Tuple[str, int]:
        assert self._server is not None, "Make sure to call start() first"
        host, port = self._server.server_address
        return (str(host), port)

    def stop(self) -> None:
        if self._server:
            self._server.shutdown()

        self._thread.join()  # type: ignore[union-attr]


@mock_aws
class TestHTTPMessageVerification:
    MESSAGES_RECEIVED = {}

    def setup_method(self, *args):
        if not settings.TEST_DECORATOR_MODE:
            raise SkipTest("Can only be tested using decorators")
        self.server = ThreadedSnsReceiver()
        self.server.start()

    def teardown_method(self, *args):
        self.server.stop()

    def test_http_message_verification(self):
        """
        This test verifies that a message sends to a HTTP endpoint contains a valid signature + PEM.
        Steps:

        1. TEST starts a server
        2. TEST creates a topic
        3. TEST creates a HTTP subscription to that topic
        4. TEST publishes a message
        5. MOTO creates a signed message + PEM file
        6. MOTO sends a HTTP request to the server
        7. SERVER receives the message send by Moto containing the signature + link to PEM file
        8. TEST downloads the PEM file from Moto
        9. TEST verifies the signature using the PEM
        """

        conn = boto3.client("sns", region_name="us-east-1")
        topic_arn = conn.create_topic(Name=f"public_{str(uuid4())[0:6]}")["TopicArn"]

        host, port = self.server.get_host_and_port()
        conn.subscribe(
            TopicArn=topic_arn, Protocol="http", Endpoint=f"http://{host}:{port}"
        )

        conn.publish(TopicArn=topic_arn, Message="mymsg", Subject="my汉字subject")

        msg = TestHTTPMessageVerification.MESSAGES_RECEIVED[topic_arn]

        string_to_sign = f"""Message
{msg['Message']}
MessageId
{msg['MessageId']}
Timestamp
{msg['Timestamp']}
TopicArn
{msg['TopicArn']}
Type
{msg['Type']}
"""

        signature = base64.b64decode(msg["Signature"])

        pem = self.download_pem(msg["SigningCertURL"])
        public_key = load_pem_x509_certificate(pem).public_key()

        message_sig_version = msg["SignatureVersion"]
        signature_hash = (
            hashes.SHA1() if message_sig_version == "1" else hashes.SHA256()
        )

        public_key.verify(
            signature,
            string_to_sign.encode(),
            padding=padding.PKCS1v15(),
            algorithm=signature_hash,
        )

    def download_pem(self, url: str) -> bytes:
        return requests.get(url).content



from uuid import uuid4

import boto3
import pytest
from botocore.client import ClientError

from moto import mock_aws, settings
from tests import EXAMPLE_AMI_ID


@mock_aws
def test_launch_template_create():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    resp = cli.create_launch_template(
        LaunchTemplateName=template_name,
        # the absolute minimum needed to create a template without other resources
        LaunchTemplateData={
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "test", "Value": "value"}],
                }
            ]
        },
    )

    assert "LaunchTemplate" in resp
    lt = resp["LaunchTemplate"]
    assert lt["LaunchTemplateName"] == template_name
    assert lt["DefaultVersionNumber"] == 1
    assert lt["LatestVersionNumber"] == 1

    with pytest.raises(ClientError) as ex:
        cli.create_launch_template(
            LaunchTemplateName=template_name,
            LaunchTemplateData={
                "TagSpecifications": [
                    {
                        "ResourceType": "instance",
                        "Tags": [{"Key": "test", "Value": "value"}],
                    }
                ]
            },
        )

    assert (
        str(ex.value)
        == "An error occurred (InvalidLaunchTemplateName.AlreadyExistsException) when calling the CreateLaunchTemplate operation: Launch template name already in use."
    )


@mock_aws
def test_create_launch_template__dryrun():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())

    with pytest.raises(ClientError) as exc:
        cli.create_launch_template(
            DryRun=True,
            LaunchTemplateName=template_name,
            LaunchTemplateData={"ImageId": "ami-abc123"},
            TagSpecifications=[
                {"ResourceType": "instance", "Tags": [{"Key": "key", "Value": "value"}]}
            ],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "DryRunOperation"
    assert (
        err["Message"]
        == "An error occurred (DryRunOperation) when calling the CreateLaunchTemplate operation: Request would have succeeded, but DryRun flag is set"
    )


@mock_aws
def test_describe_launch_template_versions():
    template_data = {
        "ImageId": "ami-abc123",
        "DisableApiTermination": False,
        "TagSpecifications": [
            {"ResourceType": "instance", "Tags": [{"Key": "test", "Value": "value"}]}
        ],
        "SecurityGroupIds": ["sg-1234", "sg-ab5678"],
    }

    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    create_resp = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData=template_data
    )

    # test using name
    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, Versions=["1"]
    )

    templ = resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]
    assert templ == template_data

    # test using id
    resp = cli.describe_launch_template_versions(
        LaunchTemplateId=create_resp["LaunchTemplate"]["LaunchTemplateId"],
        Versions=["1"],
    )

    templ = resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]
    assert templ == template_data

    # test using $Latest version
    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, Versions=["$Latest"]
    )

    templ = resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]
    assert templ == template_data

    # test using $Default version
    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, Versions=["$Default"]
    )

    templ = resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]
    assert templ == template_data


@mock_aws
def test_describe_launch_template_versions_by_name_when_absent():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = "foo"

    # test using name
    with pytest.raises(
        ClientError,
        match=f"The specified launch template, with template name {template_name}, does not exist",
    ):
        cli.describe_launch_template_versions(LaunchTemplateName=template_name)

    # test default response
    resp = cli.describe_launch_template_versions()
    assert resp["LaunchTemplateVersions"] == []

    # test using $Latest version
    resp = cli.describe_launch_template_versions(Versions=["$Latest"])
    assert resp["LaunchTemplateVersions"] == []


@mock_aws
def test_create_launch_template_version():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    create_resp = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    version_resp = cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    assert "LaunchTemplateVersion" in version_resp
    version = version_resp["LaunchTemplateVersion"]
    assert version["DefaultVersion"] is False
    assert (
        version["LaunchTemplateId"] == create_resp["LaunchTemplate"]["LaunchTemplateId"]
    )
    assert version["VersionDescription"] == "new ami"
    assert version["VersionNumber"] == 2


@mock_aws
def test_create_launch_template_version__dryrun():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    with pytest.raises(ClientError) as exc:
        cli.create_launch_template_version(
            DryRun=True,
            LaunchTemplateName=template_name,
            LaunchTemplateData={"ImageId": "ami-def456"},
            VersionDescription="new ami",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "DryRunOperation"
    assert (
        err["Message"]
        == "An error occurred (DryRunOperation) when calling the CreateLaunchTemplateVersion operation: Request would have succeeded, but DryRun flag is set"
    )


@mock_aws
def test_create_launch_template_version_by_id():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    create_resp = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    version_resp = cli.create_launch_template_version(
        LaunchTemplateId=create_resp["LaunchTemplate"]["LaunchTemplateId"],
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    assert "LaunchTemplateVersion" in version_resp
    version = version_resp["LaunchTemplateVersion"]
    assert version["DefaultVersion"] is False
    assert (
        version["LaunchTemplateId"] == create_resp["LaunchTemplate"]["LaunchTemplateId"]
    )
    assert version["VersionDescription"] == "new ami"
    assert version["VersionNumber"] == 2


@mock_aws
def test_describe_launch_template_versions_with_multiple_versions():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    resp = cli.describe_launch_template_versions(LaunchTemplateName=template_name)

    assert len(resp["LaunchTemplateVersions"]) == 2
    assert (
        resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]["ImageId"]
        == "ami-abc123"
    )
    assert (
        resp["LaunchTemplateVersions"][1]["LaunchTemplateData"]["ImageId"]
        == "ami-def456"
    )


@mock_aws
def test_describe_launch_template_versions_with_versions_option():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-hij789"},
        VersionDescription="new ami, again",
    )

    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, Versions=["2", "3"]
    )

    assert len(resp["LaunchTemplateVersions"]) == 2
    assert (
        resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]["ImageId"]
        == "ami-def456"
    )
    assert (
        resp["LaunchTemplateVersions"][1]["LaunchTemplateData"]["ImageId"]
        == "ami-hij789"
    )


@mock_aws
def test_describe_launch_template_versions_with_min():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-hij789"},
        VersionDescription="new ami, again",
    )

    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, MinVersion="2"
    )

    assert len(resp["LaunchTemplateVersions"]) == 2
    assert (
        resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]["ImageId"]
        == "ami-def456"
    )
    assert (
        resp["LaunchTemplateVersions"][1]["LaunchTemplateData"]["ImageId"]
        == "ami-hij789"
    )


@mock_aws
def test_describe_launch_template_versions_with_max():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-hij789"},
        VersionDescription="new ami, again",
    )

    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, MaxVersion="2"
    )

    assert len(resp["LaunchTemplateVersions"]) == 2
    assert (
        resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]["ImageId"]
        == "ami-abc123"
    )
    assert (
        resp["LaunchTemplateVersions"][1]["LaunchTemplateData"]["ImageId"]
        == "ami-def456"
    )


@mock_aws
def test_describe_launch_template_versions_with_min_and_max():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-def456"},
        VersionDescription="new ami",
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-hij789"},
        VersionDescription="new ami, again",
    )

    cli.create_launch_template_version(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-345abc"},
        VersionDescription="new ami, because why not",
    )

    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, MinVersion="2", MaxVersion="3"
    )

    assert len(resp["LaunchTemplateVersions"]) == 2
    assert (
        resp["LaunchTemplateVersions"][0]["LaunchTemplateData"]["ImageId"]
        == "ami-def456"
    )
    assert (
        resp["LaunchTemplateVersions"][1]["LaunchTemplateData"]["ImageId"]
        == "ami-hij789"
    )


@mock_aws
def test_describe_launch_templates_with_non_existent_name():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())

    with pytest.raises(ClientError) as ex:
        cli.describe_launch_templates(LaunchTemplateNames=[template_name])

    assert (
        str(ex.value)
        == "An error occurred (InvalidLaunchTemplateName.NotFoundException) when calling the DescribeLaunchTemplates operation: At least one of the launch templates specified in the request does not exist."
    )


@mock_aws
def test_describe_launch_templates():
    cli = boto3.client("ec2", region_name="us-east-1")

    lt_ids = []
    template_name = str(uuid4())
    r = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )
    lt_ids.append(r["LaunchTemplate"]["LaunchTemplateId"])

    template_name2 = str(uuid4())
    r = cli.create_launch_template(
        LaunchTemplateName=template_name2, LaunchTemplateData={"ImageId": "ami-abc123"}
    )
    lt_ids.append(r["LaunchTemplate"]["LaunchTemplateId"])

    # general call, all templates
    if not settings.TEST_SERVER_MODE:
        # Bug: Only 15 launch templates are returned, ever
        # ServerMode may have many more templates, created in parallel
        all_templates = retrieve_all_templates(cli)
        my_templates = [t for t in all_templates if t["LaunchTemplateId"] in lt_ids]
        assert my_templates[0]["LaunchTemplateName"] == template_name
        assert my_templates[1]["LaunchTemplateName"] == template_name2

    # filter by names
    resp = cli.describe_launch_templates(
        LaunchTemplateNames=[template_name2, template_name]
    )
    assert "LaunchTemplates" in resp
    assert len(resp["LaunchTemplates"]) == 2
    assert resp["LaunchTemplates"][0]["LaunchTemplateName"] == template_name2
    assert resp["LaunchTemplates"][1]["LaunchTemplateName"] == template_name

    # filter by ids
    resp = cli.describe_launch_templates(LaunchTemplateIds=lt_ids)
    assert "LaunchTemplates" in resp
    assert len(resp["LaunchTemplates"]) == 2
    assert resp["LaunchTemplates"][0]["LaunchTemplateName"] == template_name
    assert resp["LaunchTemplates"][1]["LaunchTemplateName"] == template_name2


@mock_aws
def test_describe_launch_templates_with_filters():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    r = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    tag_value = str(uuid4())
    cli.create_tags(
        Resources=[r["LaunchTemplate"]["LaunchTemplateId"]],
        Tags=[
            {"Key": "tag1", "Value": tag_value},
            {"Key": "another-key", "Value": "this value"},
        ],
    )

    template_name_no_tags = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name_no_tags,
        LaunchTemplateData={"ImageId": "ami-abc123"},
    )

    resp = cli.describe_launch_templates(
        Filters=[{"Name": "tag:tag1", "Values": [tag_value]}]
    )

    assert len(resp["LaunchTemplates"]) == 1
    assert resp["LaunchTemplates"][0]["LaunchTemplateName"] == template_name

    resp = cli.describe_launch_templates(
        Filters=[{"Name": "launch-template-name", "Values": [template_name_no_tags]}]
    )
    assert len(resp["LaunchTemplates"]) == 1
    assert resp["LaunchTemplates"][0]["LaunchTemplateName"] == template_name_no_tags


@mock_aws
def test_create_launch_template_with_tag_spec():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-abc123"},
        TagSpecifications=[
            {"ResourceType": "instance", "Tags": [{"Key": "key", "Value": "value"}]}
        ],
    )

    resp = cli.describe_launch_template_versions(
        LaunchTemplateName=template_name, Versions=["1"]
    )
    version = resp["LaunchTemplateVersions"][0]

    assert "TagSpecifications" in version["LaunchTemplateData"]
    assert len(version["LaunchTemplateData"]["TagSpecifications"]) == 1
    assert version["LaunchTemplateData"]["TagSpecifications"][0] == {
        "ResourceType": "instance",
        "Tags": [{"Key": "key", "Value": "value"}],
    }


@mock_aws
def test_get_launch_template_data():
    client = boto3.client("ec2", region_name="us-east-1")

    reservation = client.run_instances(ImageId=EXAMPLE_AMI_ID, MinCount=1, MaxCount=1)
    instance = reservation["Instances"][0]

    launch_template_data = client.get_launch_template_data(
        InstanceId=instance["InstanceId"]
    )["LaunchTemplateData"]

    # Ensure launch template data matches instance
    assert launch_template_data["ImageId"] == instance["ImageId"]
    assert launch_template_data["InstanceType"] == instance["InstanceType"]

    # Ensure a launch template can be created from this data
    client.create_launch_template(
        LaunchTemplateName=str(uuid4()),
        LaunchTemplateData=launch_template_data,
    )


@mock_aws
def test_delete_launch_template__dryrun():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-abc123"},
        TagSpecifications=[
            {"ResourceType": "instance", "Tags": [{"Key": "key", "Value": "value"}]}
        ],
    )

    assert (
        len(
            cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
                "LaunchTemplates"
            ]
        )
        == 1
    )

    with pytest.raises(ClientError) as exc:
        cli.delete_launch_template(DryRun=True, LaunchTemplateName=template_name)
    err = exc.value.response["Error"]
    assert err["Code"] == "DryRunOperation"
    assert (
        err["Message"]
        == "An error occurred (DryRunOperation) when calling the DeleteLaunchTemplate operation: Request would have succeeded, but DryRun flag is set"
    )

    # Template still exists
    assert (
        len(
            cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
                "LaunchTemplates"
            ]
        )
        == 1
    )


@mock_aws
def test_delete_launch_template__by_name():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )

    assert (
        len(
            cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
                "LaunchTemplates"
            ]
        )
        == 1
    )

    cli.delete_launch_template(LaunchTemplateName=template_name)

    with pytest.raises(ClientError) as exc:
        cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
            "LaunchTemplates"
        ]
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidLaunchTemplateName.NotFoundException"
    assert (
        err["Message"]
        == "At least one of the launch templates specified in the request does not exist."
    )

    # Ensure deleted template name can be reused
    cli.create_launch_template(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-abc123"},
    )


@mock_aws
def test_delete_launch_template__by_id():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())

    with pytest.raises(ClientError) as exc:
        cli.delete_launch_template()
    err = exc.value.response["Error"]
    assert err["Code"] == "MissingParameter"
    assert (
        err["Message"]
        == "The request must contain the parameter launch template ID or launch template name"
    )

    template_id = cli.create_launch_template(
        LaunchTemplateName=template_name, LaunchTemplateData={"ImageId": "ami-abc123"}
    )["LaunchTemplate"]["LaunchTemplateId"]

    assert (
        len(
            cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
                "LaunchTemplates"
            ]
        )
        == 1
    )

    cli.delete_launch_template(LaunchTemplateId=template_id)

    with pytest.raises(ClientError) as exc:
        cli.describe_launch_templates(LaunchTemplateNames=[template_name])
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidLaunchTemplateName.NotFoundException"
    assert (
        err["Message"]
        == "At least one of the launch templates specified in the request does not exist."
    )

    # Ensure deleted template name can be reused
    cli.create_launch_template(
        LaunchTemplateName=template_name,
        LaunchTemplateData={"ImageId": "ami-abc123"},
    )


def retrieve_all_templates(client, filters=[]):  # pylint: disable=W0102
    resp = client.describe_launch_templates(Filters=filters)
    all_templates = resp["LaunchTemplates"]
    next_token = resp.get("NextToken")
    while next_token:
        resp = client.describe_launch_templates(Filters=filters, NextToken=next_token)
        all_templates.extend(resp["LaunchTemplates"])
        next_token = resp.get("NextToken")
    return all_templates


@mock_aws
def test_launch_template_create_with_tags():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    resp = cli.create_launch_template(
        LaunchTemplateName=template_name,
        # the absolute minimum needed to create a template without other resources
        LaunchTemplateData={
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "test", "Value": "value"}],
                }
            ]
        },
        TagSpecifications=[
            {
                "ResourceType": "launch-template",
                "Tags": [{"Key": "test1", "Value": "value1"}],
            }
        ],
    )

    assert "LaunchTemplate" in resp
    lt = resp["LaunchTemplate"]
    assert lt["LaunchTemplateName"] == template_name
    assert lt["DefaultVersionNumber"] == 1
    assert lt["LatestVersionNumber"] == 1
    assert len(lt["Tags"]) == 1
    assert lt["Tags"][0] == {"Key": "test1", "Value": "value1"}


@mock_aws
def test_launch_template_describe_with_tags():
    cli = boto3.client("ec2", region_name="us-east-1")

    template_name = str(uuid4())
    cli.create_launch_template(
        LaunchTemplateName=template_name,
        # the absolute minimum needed to create a template without other resources
        LaunchTemplateData={
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "test", "Value": "value"}],
                }
            ]
        },
        TagSpecifications=[
            {
                "ResourceType": "launch-template",
                "Tags": [{"Key": "test1", "Value": "value1"}],
            }
        ],
    )

    lt = cli.describe_launch_templates(LaunchTemplateNames=[template_name])[
        "LaunchTemplates"
    ][0]

    assert lt["LaunchTemplateName"] == template_name
    assert lt["DefaultVersionNumber"] == 1
    assert lt["LatestVersionNumber"] == 1
    assert len(lt["Tags"]) == 1
    assert lt["Tags"][0] == {"Key": "test1", "Value": "value1"}



from __future__ import annotations

import abc
import json
import threading
from typing import Any, Final, List, Optional

from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.flow.start_at import StartAt
from moto.stepfunctions.parser.asl.component.program.program import Program
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_component import (
    IterationComponent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iteration_worker import (
    IterationWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    JobClosed,
    JobPool,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.max_concurrency import (
    DEFAULT_MAX_CONCURRENCY_VALUE,
)
from moto.stepfunctions.parser.asl.component.states import States
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.utils import TMP_THREADS


class InlineIterationComponentEvalInput:
    state_name: Final[str]
    max_concurrency: Final[int]
    input_items: Final[List[json]]

    def __init__(self, state_name: str, max_concurrency: int, input_items: List[json]):
        self.state_name = state_name
        self.max_concurrency = max_concurrency
        self.input_items = input_items


class InlineIterationComponent(IterationComponent, abc.ABC):
    _eval_input: Optional[InlineIterationComponentEvalInput]
    _job_pool: Optional[JobPool]

    def __init__(
        self,
        start_at: StartAt,
        states: States,
        comment: Optional[Comment],
    ):
        super().__init__(start_at=start_at, states=states, comment=comment)
        self._eval_input = None
        self._job_pool = None

    @abc.abstractmethod
    def _create_worker(self, env: Environment) -> IterationWorker: ...

    def _launch_worker(self, env: Environment) -> IterationWorker:
        worker = self._create_worker(env=env)
        worker_thread = threading.Thread(target=worker.eval)
        TMP_THREADS.append(worker_thread)
        worker_thread.start()
        return worker

    def _eval_body(self, env: Environment) -> None:
        self._eval_input = env.stack.pop()

        max_concurrency: int = self._eval_input.max_concurrency
        input_items: List[json] = self._eval_input.input_items

        input_item_prog: Final[Program] = Program(
            start_at=self._start_at,
            states=self._states,
            timeout_seconds=None,
            comment=self._comment,
        )
        self._job_pool = JobPool(
            job_program=input_item_prog, job_inputs=self._eval_input.input_items
        )

        number_of_workers = (
            len(input_items)
            if max_concurrency == DEFAULT_MAX_CONCURRENCY_VALUE
            else max_concurrency
        )
        for _ in range(number_of_workers):
            self._launch_worker(env=env)

        self._job_pool.await_jobs()

        worker_exception: Optional[Exception] = self._job_pool.get_worker_exception()
        if worker_exception is not None:
            raise worker_exception

        closed_jobs: List[JobClosed] = self._job_pool.get_closed_jobs()
        outputs: List[Any] = [closed_job.job_output for closed_job in closed_jobs]

        env.stack.append(outputs)



from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from tests import EXAMPLE_AMI_ID


@mock_aws
def test_route_tables_defaults():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    all_route_tables = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"]
    assert len(all_route_tables) == 1

    main_route_table = all_route_tables[0]
    assert main_route_table["VpcId"] == vpc.id

    routes = main_route_table["Routes"]
    assert len(routes) == 1

    local_route = routes[0]
    assert local_route["GatewayId"] == "local"
    assert local_route["State"] == "active"
    assert local_route["DestinationCidrBlock"] == vpc.cidr_block

    vpc.delete()

    all_route_tables = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"]
    assert len(all_route_tables) == 0


@mock_aws
def test_route_tables_additional():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    route_table = vpc.create_route_table()

    all_route_tables = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"]
    assert len(all_route_tables) == 2
    assert all_route_tables[0]["VpcId"] == vpc.id
    assert all_route_tables[1]["VpcId"] == vpc.id

    all_route_table_ids = [r["RouteTableId"] for r in all_route_tables]
    assert route_table.route_table_id in all_route_table_ids

    routes = route_table.routes
    assert len(routes) == 1

    local_route = routes[0]
    assert local_route.gateway_id == "local"
    assert local_route.state == "active"
    assert local_route.destination_cidr_block == vpc.cidr_block

    with pytest.raises(ClientError) as ex:
        client.delete_vpc(VpcId=vpc.id)
    assert ex.value.response["Error"]["Code"] == "DependencyViolation"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]

    client.delete_route_table(RouteTableId=route_table.route_table_id)

    all_route_tables = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"]
    assert len(all_route_tables) == 1

    with pytest.raises(ClientError) as ex:
        client.delete_route_table(RouteTableId="rtb-1234abcd")
    assert ex.value.response["Error"]["Code"] == "InvalidRouteTableID.NotFound"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]


@mock_aws
def test_route_tables_filters_standard():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")

    vpc1 = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    route_table1 = ec2.create_route_table(VpcId=vpc1.id)

    vpc2 = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    route_table2 = ec2.create_route_table(VpcId=vpc2.id)
    igw = ec2.create_internet_gateway()
    route_table2.create_route(DestinationCidrBlock="10.0.0.4/24", GatewayId=igw.id)

    all_route_tables = client.describe_route_tables()["RouteTables"]
    all_ids = [rt["RouteTableId"] for rt in all_route_tables]
    assert route_table1.id in all_ids
    assert route_table2.id in all_ids

    # Filter by main route table
    main_route_tables = client.describe_route_tables(
        Filters=[{"Name": "association.main", "Values": ["true"]}]
    )["RouteTables"]
    main_route_table_ids = [
        route_table["RouteTableId"] for route_table in main_route_tables
    ]
    assert route_table1.id not in main_route_table_ids
    assert route_table2.id not in main_route_table_ids

    # Filter by VPC
    vpc1_route_tables = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc1.id]}]
    )["RouteTables"]
    assert len(vpc1_route_tables) == 2
    vpc1_route_table_ids = [
        route_table["RouteTableId"] for route_table in vpc1_route_tables
    ]
    assert route_table1.id in vpc1_route_table_ids
    assert route_table2.id not in vpc1_route_table_ids

    # Filter by VPC and main route table
    vpc2_main_route_tables = client.describe_route_tables(
        Filters=[
            {"Name": "association.main", "Values": ["true"]},
            {"Name": "vpc-id", "Values": [vpc2.id]},
        ]
    )["RouteTables"]
    assert len(vpc2_main_route_tables) == 1
    vpc2_main_route_table_ids = [
        route_table["RouteTableId"] for route_table in vpc2_main_route_tables
    ]
    assert route_table1.id not in vpc2_main_route_table_ids
    assert route_table2.id not in vpc2_main_route_table_ids

    # Filter by route gateway id
    resp = client.describe_route_tables(
        Filters=[
            {"Name": "route.gateway-id", "Values": [igw.id]},
        ]
    )["RouteTables"]
    assert any(
        [route["GatewayId"] == igw.id for table in resp for route in table["Routes"]]
    )

    # Filter by route destination CIDR block
    resp = client.describe_route_tables(
        Filters=[
            {"Name": "route.destination-cidr-block", "Values": ["10.0.0.4/24"]},
        ]
    )["RouteTables"]
    assert any([route_table["RouteTableId"] == route_table2.id for route_table in resp])
    assert any(
        [
            route["DestinationCidrBlock"] == "10.0.0.4/24"
            for table in resp
            for route in table["Routes"]
        ]
    )

    # Unsupported filter
    if not settings.TEST_SERVER_MODE:
        # ServerMode will just throw a generic 500
        filters = [{"Name": "not-implemented-filter", "Values": ["foobar"]}]
        with pytest.raises(NotImplementedError):
            client.describe_route_tables(Filters=filters)


@mock_aws
def test_route_tables_filters_associations():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet1 = vpc.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/24")
    subnet2 = vpc.create_subnet(VpcId=vpc.id, CidrBlock="10.0.1.0/24")
    subnet3 = vpc.create_subnet(VpcId=vpc.id, CidrBlock="10.0.2.0/24")
    route_table1 = ec2.create_route_table(VpcId=vpc.id)
    route_table2 = ec2.create_route_table(VpcId=vpc.id)

    association_id1 = client.associate_route_table(
        RouteTableId=route_table1.id, SubnetId=subnet1.id
    )["AssociationId"]
    client.associate_route_table(RouteTableId=route_table1.id, SubnetId=subnet2.id)
    client.associate_route_table(RouteTableId=route_table2.id, SubnetId=subnet3.id)

    # Filter by association ID
    association1_route_tables = client.describe_route_tables(
        Filters=[
            {
                "Name": "association.route-table-association-id",
                "Values": [association_id1],
            }
        ]
    )["RouteTables"]
    assert len(association1_route_tables) == 1
    assert association1_route_tables[0]["RouteTableId"] == route_table1.id
    assert len(association1_route_tables[0]["Associations"]) == 2

    # Filter by route table ID
    route_table2_route_tables = client.describe_route_tables(
        Filters=[{"Name": "association.route-table-id", "Values": [route_table2.id]}]
    )["RouteTables"]
    assert len(route_table2_route_tables) == 1
    assert route_table2_route_tables[0]["RouteTableId"] == route_table2.id
    assert len(route_table2_route_tables[0]["Associations"]) == 1

    # Filter by subnet ID
    subnet_route_tables = client.describe_route_tables(
        Filters=[{"Name": "association.subnet-id", "Values": [subnet1.id]}]
    )["RouteTables"]
    assert len(subnet_route_tables) == 1
    assert subnet_route_tables[0]["RouteTableId"] == route_table1.id
    assert len(subnet_route_tables[0]["Associations"]) == 2


@mock_aws
def test_route_tables_filters_vpc_peering_connection():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    main_route_table_id = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)
    ROUTE_CIDR = "10.0.0.4/24"

    peer_vpc = ec2.create_vpc(CidrBlock="11.0.0.0/16")
    vpc_pcx = ec2.create_vpc_peering_connection(VpcId=vpc.id, PeerVpcId=peer_vpc.id)

    main_route_table.create_route(
        DestinationCidrBlock=ROUTE_CIDR, VpcPeeringConnectionId=vpc_pcx.id
    )

    # Refresh route table
    main_route_table.reload()
    new_routes = [
        route
        for route in main_route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 1

    new_route = new_routes[0]
    assert new_route.gateway_id is None
    assert new_route.instance_id is None
    assert new_route.vpc_peering_connection_id == vpc_pcx.id
    assert new_route.state == "active"
    assert new_route.destination_cidr_block == ROUTE_CIDR

    # Filter by Peering Connection
    route_tables = client.describe_route_tables(
        Filters=[{"Name": "route.vpc-peering-connection-id", "Values": [vpc_pcx.id]}]
    )["RouteTables"]
    assert len(route_tables) == 1
    route_table = route_tables[0]
    assert route_table["RouteTableId"] == main_route_table_id
    vpc_pcx_ids = [
        route["VpcPeeringConnectionId"]
        for route in route_table["Routes"]
        if "VpcPeeringConnectionId" in route
    ]
    all(vpc_pcx_id == vpc_pcx.id for vpc_pcx_id in vpc_pcx_ids)


@mock_aws
def test_route_tables_filters_transit_gateway():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    main_route_table_id = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)

    response = client.create_transit_gateway()

    gateway = response["TransitGateway"]
    main_route_table.create_route(TransitGatewayId=gateway["TransitGatewayId"])
    route_tables = client.describe_route_tables(
        Filters=[
            {
                "Name": "route.transit-gateway-id",
                "Values": [gateway["TransitGatewayId"]],
            }
        ]
    )["RouteTables"]

    assert len(route_tables) == 1
    route_table = route_tables[0]
    assert route_table["RouteTableId"] == main_route_table_id

    # Filter using regex
    route_tables = client.describe_route_tables(
        Filters=[
            {
                "Name": "route.transit-gateway-id",
                "Values": ["tgw-*"],
            }
        ]
    )["RouteTables"]

    assert len(route_tables) == 1
    route_table = route_tables[0]
    assert route_table["RouteTableId"] == main_route_table_id


@mock_aws
def test_route_table_associations():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/18")
    route_table = ec2.create_route_table(VpcId=vpc.id)

    # Refresh
    r = client.describe_route_tables(RouteTableIds=[route_table.id])["RouteTables"][0]
    assert len(r["Associations"]) == 0

    # Associate
    association_id = client.associate_route_table(
        RouteTableId=route_table.id, SubnetId=subnet.id
    )["AssociationId"]

    # Refresh
    r = client.describe_route_tables(RouteTableIds=[route_table.id])["RouteTables"][0]
    assert len(r["Associations"]) == 1

    assert r["Associations"][0]["RouteTableAssociationId"] == association_id
    assert r["Associations"][0]["Main"] is False
    assert r["Associations"][0]["RouteTableId"] == route_table.id
    assert r["Associations"][0]["SubnetId"] == subnet.id

    # Associate is idempotent
    association_id_idempotent = client.associate_route_table(
        RouteTableId=route_table.id, SubnetId=subnet.id
    )["AssociationId"]
    assert association_id_idempotent == association_id

    # Error: Attempt delete associated route table.
    with pytest.raises(ClientError) as ex:
        client.delete_route_table(RouteTableId=route_table.id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "DependencyViolation"

    # Disassociate
    client.disassociate_route_table(AssociationId=association_id)

    # Validate
    r = client.describe_route_tables(RouteTableIds=[route_table.id])["RouteTables"][0]
    assert len(r["Associations"]) == 0

    # Error: Disassociate with invalid association ID
    with pytest.raises(ClientError) as ex:
        client.disassociate_route_table(AssociationId=association_id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidAssociationID.NotFound"

    # Error: Associate with invalid subnet ID
    with pytest.raises(ClientError) as ex:
        client.associate_route_table(
            RouteTableId=route_table.id, SubnetId="subnet-1234abcd"
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidSubnetID.NotFound"

    # Error: Associate with invalid route table ID
    with pytest.raises(ClientError) as ex:
        client.associate_route_table(RouteTableId="rtb-1234abcd", SubnetId=subnet.id)
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidRouteTableID.NotFound"


@mock_aws
def test_route_table_replace_route_table_association():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/18")
    route_table1_id = ec2.create_route_table(VpcId=vpc.id).id
    route_table2_id = ec2.create_route_table(VpcId=vpc.id).id

    all_route_tables = client.describe_route_tables()["RouteTables"]
    all_ids = [rt["RouteTableId"] for rt in all_route_tables]
    assert route_table1_id in all_ids
    assert route_table2_id in all_ids

    # Refresh
    route_table1 = client.describe_route_tables(RouteTableIds=[route_table1_id])[
        "RouteTables"
    ][0]
    assert len(route_table1["Associations"]) == 0

    # Associate
    association_id1 = client.associate_route_table(
        RouteTableId=route_table1_id, SubnetId=subnet.id
    )["AssociationId"]

    # Refresh
    route_table1 = client.describe_route_tables(RouteTableIds=[route_table1_id])[
        "RouteTables"
    ][0]
    route_table2 = client.describe_route_tables(RouteTableIds=[route_table2_id])[
        "RouteTables"
    ][0]

    # Validate
    assert len(route_table1["Associations"]) == 1
    assert len(route_table2["Associations"]) == 0

    assert route_table1["Associations"][0]["RouteTableAssociationId"] == association_id1
    assert route_table1["Associations"][0]["Main"] is False
    assert route_table1["Associations"][0]["RouteTableId"] == route_table1_id
    assert route_table1["Associations"][0]["SubnetId"] == subnet.id

    # Replace Association
    association_id2 = client.replace_route_table_association(
        AssociationId=association_id1, RouteTableId=route_table2_id
    )["NewAssociationId"]

    # Refresh
    route_table1 = client.describe_route_tables(RouteTableIds=[route_table1_id])[
        "RouteTables"
    ][0]
    route_table2 = client.describe_route_tables(RouteTableIds=[route_table2_id])[
        "RouteTables"
    ][0]

    # Validate
    assert len(route_table1["Associations"]) == 0
    assert len(route_table2["Associations"]) == 1

    assert route_table2["Associations"][0]["RouteTableAssociationId"] == association_id2
    assert route_table2["Associations"][0]["Main"] is False
    assert route_table2["Associations"][0]["RouteTableId"] == route_table2_id
    assert route_table2["Associations"][0]["SubnetId"] == subnet.id

    # Replace Association is idempotent
    association_id_idempotent = client.replace_route_table_association(
        AssociationId=association_id2, RouteTableId=route_table2_id
    )["NewAssociationId"]
    assert association_id_idempotent == association_id2

    # Error: Replace association with invalid association ID
    with pytest.raises(ClientError) as ex:
        client.replace_route_table_association(
            AssociationId="rtbassoc-1234abcd", RouteTableId=route_table1_id
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidAssociationID.NotFound"

    # Error: Replace association with invalid route table ID
    with pytest.raises(ClientError) as ex:
        client.replace_route_table_association(
            AssociationId=association_id2, RouteTableId="rtb-1234abcd"
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidRouteTableID.NotFound"


@mock_aws
def test_route_table_replace_route_table_association_for_main():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    new_route_table_id = ec2.create_route_table(VpcId=vpc.id).id

    # Get main route table details
    main_route_table = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]
    main_route_table_id = main_route_table["RouteTableId"]
    main_route_table_association_id = main_route_table["Associations"][0][
        "RouteTableAssociationId"
    ]

    # Replace Association
    new_association = client.replace_route_table_association(
        AssociationId=main_route_table_association_id, RouteTableId=new_route_table_id
    )
    new_association_id = new_association["NewAssociationId"]

    # Validate the format
    assert new_association["AssociationState"]["State"] == "associated"

    # Refresh
    main_route_table = client.describe_route_tables(
        RouteTableIds=[main_route_table_id]
    )["RouteTables"][0]
    new_route_table = client.describe_route_tables(RouteTableIds=[new_route_table_id])[
        "RouteTables"
    ][0]

    # Validate
    assert len(main_route_table["Associations"]) == 0
    assert len(new_route_table["Associations"]) == 1
    assert (
        new_route_table["Associations"][0]["RouteTableAssociationId"]
        == new_association_id
    )
    assert new_route_table["Associations"][0]["Main"] is True


@mock_aws
def test_route_table_get_by_tag():
    ec2 = boto3.resource("ec2", region_name="eu-central-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    route_table = ec2.create_route_table(VpcId=vpc.id)
    tag_value = str(uuid4())
    route_table.create_tags(Tags=[{"Key": "Name", "Value": tag_value}])

    filters = [{"Name": "tag:Name", "Values": [tag_value]}]
    route_tables = list(ec2.route_tables.filter(Filters=filters))

    assert len(route_tables) == 1
    assert route_tables[0].vpc_id == vpc.id
    assert route_tables[0].id == route_table.id
    assert len(route_tables[0].tags) == 1
    assert route_tables[0].tags[0] == {"Key": "Name", "Value": tag_value}


@mock_aws
def test_routes_additional():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    main_route_table_id = client.describe_route_tables(
        Filters=[{"Name": "vpc-id", "Values": [vpc.id]}]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)

    assert len(main_route_table.routes) == 1
    igw = ec2.create_internet_gateway()
    ROUTE_CIDR = "10.0.0.4/24"

    main_route_table.create_route(DestinationCidrBlock=ROUTE_CIDR, GatewayId=igw.id)

    assert len(main_route_table.routes) == 2
    new_routes = [
        route
        for route in main_route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 1

    new_route = new_routes[0]
    assert new_route.gateway_id == igw.id
    assert new_route.instance_id is None
    assert new_route.state == "active"
    assert new_route.destination_cidr_block == ROUTE_CIDR

    client.delete_route(
        RouteTableId=main_route_table.id, DestinationCidrBlock=ROUTE_CIDR
    )

    main_route_table.reload()

    assert len(main_route_table.routes) == 1
    new_routes = [
        route
        for route in main_route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 0

    with pytest.raises(ClientError) as ex:
        client.delete_route(
            RouteTableId=main_route_table.id, DestinationCidrBlock=ROUTE_CIDR
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidRoute.NotFound"


@mock_aws
def test_routes_replace():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/24")

    main_route_table_id = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)
    ROUTE_CIDR = "10.0.0.4/24"

    # Various route targets
    igw = ec2.create_internet_gateway()

    instance = ec2.create_instances(ImageId=EXAMPLE_AMI_ID, MinCount=1, MaxCount=1)[0]

    eni = ec2.create_network_interface(SubnetId=subnet.id)

    # Create initial route
    main_route_table.create_route(DestinationCidrBlock=ROUTE_CIDR, GatewayId=igw.id)

    # Replace...
    def get_target_route():
        route_table = client.describe_route_tables(RouteTableIds=[main_route_table.id])[
            "RouteTables"
        ][0]
        routes = [
            route
            for route in route_table["Routes"]
            if route["DestinationCidrBlock"] != vpc.cidr_block
        ]
        assert len(routes) == 1
        return routes[0]

    client.replace_route(
        RouteTableId=main_route_table.id,
        DestinationCidrBlock=ROUTE_CIDR,
        InstanceId=instance.id,
    )

    target_route = get_target_route()
    assert "GatewayId" not in target_route
    assert target_route["InstanceId"] == instance.id
    assert "NetworkInterfaceId" not in target_route
    assert target_route["State"] == "active"
    assert target_route["DestinationCidrBlock"] == ROUTE_CIDR

    client.replace_route(
        RouteTableId=main_route_table.id,
        DestinationCidrBlock=ROUTE_CIDR,
        GatewayId=igw.id,
    )

    target_route = get_target_route()
    assert target_route["GatewayId"] == igw.id
    assert "InstanceId" not in target_route
    assert "NetworkInterfaceId" not in target_route
    assert target_route["State"] == "active"
    assert target_route["DestinationCidrBlock"] == ROUTE_CIDR

    client.replace_route(
        RouteTableId=main_route_table.id,
        DestinationCidrBlock=ROUTE_CIDR,
        NetworkInterfaceId=eni.id,
    )

    target_route = get_target_route()
    assert "GatewayId" not in target_route
    assert "InstanceId" not in target_route
    assert target_route["NetworkInterfaceId"] == eni.id
    assert target_route["State"] == "active"
    assert target_route["DestinationCidrBlock"] == ROUTE_CIDR

    with pytest.raises(ClientError) as ex:
        client.replace_route(
            RouteTableId="rtb-1234abcd",
            DestinationCidrBlock=ROUTE_CIDR,
            GatewayId=igw.id,
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidRouteTableID.NotFound"

    with pytest.raises(ClientError) as ex:
        client.replace_route(
            RouteTableId=main_route_table.id,
            DestinationCidrBlock="1.1.1.1/32",
            NetworkInterfaceId=eni.id,
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    # This should be 'InvalidRoute.NotFound' in line with the delete_route()
    # equivalent, but for some reason AWS returns InvalidParameterValue instead.
    assert ex.value.response["Error"]["Code"] == "InvalidParameterValue"


@mock_aws
def test_routes_already_exist():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    main_route_table_id = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)
    ROUTE_CIDR = "10.0.0.0/23"
    ROUTE_SUB_CIDR = "10.0.0.0/24"
    ROUTE_NO_CONFLICT_CIDR = "10.0.2.0/24"

    # Various route targets
    igw = ec2.create_internet_gateway()

    # Create initial route
    main_route_table.create_route(DestinationCidrBlock=ROUTE_CIDR, GatewayId=igw.id)
    main_route_table.create_route(
        DestinationCidrBlock=ROUTE_NO_CONFLICT_CIDR, GatewayId=igw.id
    )

    # Create
    with pytest.raises(ClientError) as ex:
        client.create_route(
            RouteTableId=main_route_table.id,
            DestinationCidrBlock=ROUTE_CIDR,
            GatewayId=igw.id,
        )

    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "RouteAlreadyExists"

    # We can create a sub cidr
    client.create_route(
        RouteTableId=main_route_table.id,
        DestinationCidrBlock=ROUTE_SUB_CIDR,
        GatewayId=igw.id,
    )
    # Or even a catch-all
    client.create_route(
        RouteTableId=main_route_table.id,
        DestinationCidrBlock="0.0.0.0/0",
        GatewayId=igw.id,
    )


@mock_aws
def test_routes_not_supported():
    client = boto3.client("ec2", region_name="us-east-1")
    main_route_table_id = client.describe_route_tables()["RouteTables"][0][
        "RouteTableId"
    ]

    ROUTE_CIDR = "10.0.0.4/24"

    # Create
    with pytest.raises(ClientError) as ex:
        client.create_route(
            RouteTableId=main_route_table_id,
            DestinationCidrBlock=ROUTE_CIDR,
            NetworkInterfaceId="eni-1234abcd",
        )
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidNetworkInterfaceID.NotFound"


@mock_aws
def test_routes_vpc_peering_connection():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    main_route_table_id = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)
    ROUTE_CIDR = "10.0.0.4/24"

    peer_vpc = ec2.create_vpc(CidrBlock="11.0.0.0/16")
    vpc_pcx = ec2.create_vpc_peering_connection(VpcId=vpc.id, PeerVpcId=peer_vpc.id)

    main_route_table.create_route(
        DestinationCidrBlock=ROUTE_CIDR, VpcPeeringConnectionId=vpc_pcx.id
    )

    # Refresh route table
    main_route_table.reload()
    new_routes = [
        route
        for route in main_route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 1

    new_route = new_routes[0]
    assert new_route.gateway_id is None
    assert new_route.instance_id is None
    assert new_route.vpc_peering_connection_id == vpc_pcx.id
    assert new_route.state == "active"
    assert new_route.destination_cidr_block == ROUTE_CIDR


@mock_aws
def test_routes_vpn_gateway():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    main_route_table_id = client.describe_route_tables(
        Filters=[
            {"Name": "vpc-id", "Values": [vpc.id]},
            {"Name": "association.main", "Values": ["true"]},
        ]
    )["RouteTables"][0]["RouteTableId"]
    main_route_table = ec2.RouteTable(main_route_table_id)
    ROUTE_CIDR = "10.0.0.4/24"

    vpn_gw_id = client.create_vpn_gateway(Type="ipsec.1")["VpnGateway"]["VpnGatewayId"]

    main_route_table.create_route(DestinationCidrBlock=ROUTE_CIDR, GatewayId=vpn_gw_id)

    main_route_table.reload()
    new_routes = [
        route
        for route in main_route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 1

    new_route = new_routes[0]
    assert new_route.gateway_id == vpn_gw_id
    assert new_route.instance_id is None
    assert new_route.vpc_peering_connection_id is None


@mock_aws
def test_network_acl_tagging():
    client = boto3.client("ec2", region_name="us-east-1")
    ec2 = boto3.resource("ec2", region_name="us-east-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    route_table = ec2.create_route_table(VpcId=vpc.id)
    route_table.create_tags(Tags=[{"Key": "a key", "Value": "some value"}])

    tag = client.describe_tags(
        Filters=[{"Name": "resource-id", "Values": [route_table.id]}]
    )["Tags"][0]
    assert tag["ResourceType"] == "route-table"
    assert tag["Key"] == "a key"
    assert tag["Value"] == "some value"

    all_route_tables = client.describe_route_tables()["RouteTables"]
    test_route_table = next(
        na for na in all_route_tables if na["RouteTableId"] == route_table.id
    )
    assert test_route_table["Tags"] == [{"Value": "some value", "Key": "a key"}]


@mock_aws
def test_create_route_with_invalid_destination_cidr_block_parameter():
    ec2 = boto3.resource("ec2", region_name="us-west-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    vpc.reload()
    assert vpc.is_default is False

    route_table = ec2.create_route_table(VpcId=vpc.id)
    route_table.reload()

    internet_gateway = ec2.create_internet_gateway()
    vpc.attach_internet_gateway(InternetGatewayId=internet_gateway.id)
    internet_gateway.reload()

    destination_cidr_block = "1000.1.0.0/20"
    with pytest.raises(ClientError) as ex:
        route_table.create_route(
            DestinationCidrBlock=destination_cidr_block, GatewayId=internet_gateway.id
        )
    assert (
        str(ex.value)
        == f"An error occurred (InvalidParameterValue) when calling the CreateRoute operation: Value ({destination_cidr_block}) for parameter destinationCidrBlock is invalid. This is not a valid CIDR block."
    )

    route_table.create_route(
        DestinationIpv6CidrBlock="2001:db8::/125", GatewayId=internet_gateway.id
    )
    new_routes = [
        route
        for route in route_table.routes
        if route.destination_cidr_block != vpc.cidr_block
        or route.destination_ipv6_cidr_block != vpc.cidr_block
    ]
    assert len(new_routes) == 1
    assert new_routes[0].route_table_id is not None


@mock_aws
def test_create_route_with_network_interface_id():
    ec2 = boto3.resource("ec2", region_name="us-west-2")
    ec2_client = boto3.client("ec2", region_name="us-west-2")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(
        VpcId=vpc.id, CidrBlock="10.0.0.0/24", AvailabilityZone="us-west-2a"
    )

    route_table = ec2_client.create_route_table(VpcId=vpc.id)

    route_table_id = route_table["RouteTable"]["RouteTableId"]

    eni1 = ec2_client.create_network_interface(
        SubnetId=subnet.id, PrivateIpAddress="10.0.10.5"
    )

    route = ec2_client.create_route(
        NetworkInterfaceId=eni1["NetworkInterface"]["NetworkInterfaceId"],
        RouteTableId=route_table_id,
    )

    assert route["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_aws
def test_describe_route_tables_with_nat_gateway():
    ec2 = boto3.client("ec2", region_name="us-west-1")
    vpc_id = ec2.create_vpc(CidrBlock="192.168.0.0/23")["Vpc"]["VpcId"]
    igw_id = ec2.create_internet_gateway()["InternetGateway"]["InternetGatewayId"]
    ec2.attach_internet_gateway(VpcId=vpc_id, InternetGatewayId=igw_id)
    az = ec2.describe_availability_zones()["AvailabilityZones"][0]["ZoneName"]
    sn_id = ec2.create_subnet(
        AvailabilityZone=az, CidrBlock="192.168.0.0/24", VpcId=vpc_id
    )["Subnet"]["SubnetId"]
    route_table_id = ec2.create_route_table(VpcId=vpc_id)["RouteTable"]["RouteTableId"]
    ec2.associate_route_table(SubnetId=sn_id, RouteTableId=route_table_id)
    alloc_id = ec2.allocate_address(Domain="vpc")["AllocationId"]
    nat_gw_id = ec2.create_nat_gateway(SubnetId=sn_id, AllocationId=alloc_id)[
        "NatGateway"
    ]["NatGatewayId"]
    ec2.create_route(
        DestinationCidrBlock="0.0.0.0/0",
        NatGatewayId=nat_gw_id,
        RouteTableId=route_table_id,
    )

    route_table = ec2.describe_route_tables(
        Filters=[{"Name": "route-table-id", "Values": [route_table_id]}]
    )["RouteTables"][0]
    nat_gw_routes = [
        route
        for route in route_table["Routes"]
        if route["DestinationCidrBlock"] == "0.0.0.0/0"
    ]

    assert len(nat_gw_routes) == 1
    assert nat_gw_routes[0]["DestinationCidrBlock"] == "0.0.0.0/0"
    assert nat_gw_routes[0]["NatGatewayId"] == nat_gw_id
    assert nat_gw_routes[0]["State"] == "active"


@mock_aws
def test_describe_route_tables_filter_with_nat_gateway_id():
    ec2 = boto3.client("ec2", region_name="us-west-1")
    vpc_id = ec2.create_vpc(CidrBlock="192.168.0.0/23")["Vpc"]["VpcId"]
    az = ec2.describe_availability_zones()["AvailabilityZones"][0]["ZoneName"]
    sn_id = ec2.create_subnet(
        AvailabilityZone=az, CidrBlock="192.168.0.0/24", VpcId=vpc_id
    )["Subnet"]["SubnetId"]
    route_table_id = ec2.create_route_table(VpcId=vpc_id)["RouteTable"]["RouteTableId"]
    ec2.associate_route_table(SubnetId=sn_id, RouteTableId=route_table_id)
    alloc_id = ec2.allocate_address(Domain="vpc")["AllocationId"]
    nat_gw_id = ec2.create_nat_gateway(SubnetId=sn_id, AllocationId=alloc_id)[
        "NatGateway"
    ]["NatGatewayId"]

    route_tables_by_nat_gateway_id = ec2.describe_route_tables(
        Filters=[{"Name": "route.nat-gateway-id", "Values": [nat_gw_id]}]
    )["RouteTables"]
    assert len(route_tables_by_nat_gateway_id) == 0

    ec2.create_route(
        DestinationCidrBlock="0.0.0.0/0",
        NatGatewayId=nat_gw_id,
        RouteTableId=route_table_id,
    )

    route_tables_by_nat_gateway_id = ec2.describe_route_tables(
        Filters=[{"Name": "route.nat-gateway-id", "Values": [nat_gw_id]}]
    )["RouteTables"][0]
    nat_gw_routes = [
        route
        for route in route_tables_by_nat_gateway_id["Routes"]
        if route["DestinationCidrBlock"] == "0.0.0.0/0"
    ]

    assert len(nat_gw_routes) == 1
    assert nat_gw_routes[0]["DestinationCidrBlock"] == "0.0.0.0/0"
    assert nat_gw_routes[0]["NatGatewayId"] == nat_gw_id
    assert nat_gw_routes[0]["State"] == "active"


@mock_aws
def test_create_vpc_end_point():
    ec2 = boto3.client("ec2", region_name="us-west-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(VpcId=vpc["Vpc"]["VpcId"], CidrBlock="10.0.0.0/24")

    route_table = ec2.create_route_table(VpcId=vpc["Vpc"]["VpcId"])

    # test without any end point type specified
    vpc_end_point = ec2.create_vpc_endpoint(
        VpcId=vpc["Vpc"]["VpcId"],
        ServiceName="com.amazonaws.us-east-1.s3",
        RouteTableIds=[route_table["RouteTable"]["RouteTableId"]],
    )

    assert vpc_end_point["VpcEndpoint"]["ServiceName"] == "com.amazonaws.us-east-1.s3"
    assert (
        vpc_end_point["VpcEndpoint"]["RouteTableIds"][0]
        == route_table["RouteTable"]["RouteTableId"]
    )
    assert vpc_end_point["VpcEndpoint"]["VpcId"] == vpc["Vpc"]["VpcId"]
    assert len(vpc_end_point["VpcEndpoint"]["DnsEntries"]) == 0

    # test with any end point type as gateway
    vpc_end_point = ec2.create_vpc_endpoint(
        VpcId=vpc["Vpc"]["VpcId"],
        ServiceName="com.amazonaws.us-east-1.s3",
        RouteTableIds=[route_table["RouteTable"]["RouteTableId"]],
        VpcEndpointType="gateway",
    )

    assert vpc_end_point["VpcEndpoint"]["ServiceName"] == "com.amazonaws.us-east-1.s3"
    assert (
        vpc_end_point["VpcEndpoint"]["RouteTableIds"][0]
        == route_table["RouteTable"]["RouteTableId"]
    )
    assert vpc_end_point["VpcEndpoint"]["VpcId"] == vpc["Vpc"]["VpcId"]
    assert len(vpc_end_point["VpcEndpoint"]["DnsEntries"]) == 0

    # test with end point type as interface
    vpc_end_point = ec2.create_vpc_endpoint(
        VpcId=vpc["Vpc"]["VpcId"],
        ServiceName="com.amazonaws.us-east-1.s3",
        SubnetIds=[subnet["Subnet"]["SubnetId"]],
        VpcEndpointType="interface",
    )

    assert vpc_end_point["VpcEndpoint"]["ServiceName"] == "com.amazonaws.us-east-1.s3"
    assert vpc_end_point["VpcEndpoint"]["SubnetIds"][0] == subnet["Subnet"]["SubnetId"]
    assert vpc_end_point["VpcEndpoint"]["VpcId"] == vpc["Vpc"]["VpcId"]
    assert len(vpc_end_point["VpcEndpoint"]["DnsEntries"]) > 0


@mock_aws
def test_create_route_tables_with_tags():
    ec2 = boto3.resource("ec2", region_name="eu-central-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")

    route_table = ec2.create_route_table(
        VpcId=vpc.id,
        TagSpecifications=[
            {
                "ResourceType": "route-table",
                "Tags": [{"Key": "test", "Value": "TestRouteTable"}],
            }
        ],
    )

    assert len(route_table.tags) == 1


@mock_aws
def test_create_route_with_egress_only_igw():
    ec2 = boto3.resource("ec2", region_name="eu-central-1")
    ec2_client = boto3.client("ec2", region_name="eu-central-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    eigw = ec2_client.create_egress_only_internet_gateway(VpcId=vpc.id)
    eigw_id = eigw["EgressOnlyInternetGateway"]["EgressOnlyInternetGatewayId"]

    route_table = ec2.create_route_table(VpcId=vpc.id)

    ec2_client.create_route(
        RouteTableId=route_table.id,
        EgressOnlyInternetGatewayId=eigw_id,
        DestinationIpv6CidrBlock="::/0",
    )

    route_table.reload()
    eigw_route = [
        r
        for r in route_table.routes_attribute
        if r.get("DestinationIpv6CidrBlock") == "::/0"
    ][0]
    assert eigw_route.get("EgressOnlyInternetGatewayId") == eigw_id
    assert eigw_route.get("State") == "active"


@mock_aws
def test_create_route_with_unknown_egress_only_igw():
    ec2 = boto3.resource("ec2", region_name="eu-central-1")
    ec2_client = boto3.client("ec2", region_name="eu-central-1")

    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    ec2.create_subnet(
        VpcId=vpc.id, CidrBlock="10.0.0.0/24", AvailabilityZone="us-west-2a"
    )

    route_table = ec2.create_route_table(VpcId=vpc.id)

    with pytest.raises(ClientError) as ex:
        ec2_client.create_route(
            RouteTableId=route_table.id, EgressOnlyInternetGatewayId="eoigw"
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "InvalidGatewayID.NotFound"
    assert err["Message"] == "The eigw ID 'eoigw' does not exist"


@mock_aws
def test_create_route_with_vpc_endpoint():
    # Setup
    _, ec2_client, route_table, vpc = setup_vpc()
    dest_cidr = "0.0.0.0/0"
    vpc_end_point = ec2_client.create_vpc_endpoint(
        VpcId=vpc.id,
        ServiceName="com.amazonaws.vpce.eu-central-1.vpce-svc-084fa044c50cb1290",
        RouteTableIds=[route_table.id],
        VpcEndpointType="GatewayLoadBalancer",
    )
    vpce_id = vpc_end_point["VpcEndpoint"]["VpcEndpointId"]

    # Execute
    ec2_client.create_route(
        DestinationCidrBlock=dest_cidr,
        VpcEndpointId=vpce_id,
        RouteTableId=route_table.id,
    )
    rt = ec2_client.describe_route_tables(RouteTableIds=[route_table.id])
    new_route = rt["RouteTables"][-1]["Routes"][1]

    # Verify
    assert new_route["DestinationCidrBlock"] == dest_cidr
    assert new_route["GatewayId"] == vpce_id


@mock_aws
def test_create_route_with_invalid_vpc_endpoint():
    # Setup
    _, ec2_client, route_table, vpc = setup_vpc()
    dest_cidr = "0.0.0.0/0"
    vpc_end_point = ec2_client.create_vpc_endpoint(
        VpcId=vpc.id,
        ServiceName="com.amazonaws.us-east-1.s3",
        RouteTableIds=[route_table.id],
        VpcEndpointType="Gateway",
    )
    vpce_id = vpc_end_point["VpcEndpoint"]["VpcEndpointId"]

    # Execute
    with pytest.raises(ClientError) as ex:
        ec2_client.create_route(
            DestinationCidrBlock=dest_cidr,
            VpcEndpointId=vpce_id,
            RouteTableId=route_table.id,
        )
    # Verify
    err = ex.value.response["Error"]
    assert err["Code"] == "RouteNotSupported"
    assert (
        err["Message"] == f"Route table contains unsupported route target: {vpce_id}. "
        "VPC Endpoints of this type cannot be used as route targets."
    )


@mock_aws
def test_associate_route_table_by_gateway():
    ec2 = boto3.client("ec2", region_name="us-west-1")
    vpc_id = ec2.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    route_table_id = ec2.create_route_table(VpcId=vpc_id)["RouteTable"]["RouteTableId"]
    igw_id = ec2.create_internet_gateway()["InternetGateway"]["InternetGatewayId"]
    assoc_id = ec2.associate_route_table(RouteTableId=route_table_id, GatewayId=igw_id)[
        "AssociationId"
    ]
    verify = ec2.describe_route_tables(
        Filters=[
            {"Name": "association.route-table-association-id", "Values": [assoc_id]}
        ]
    )["RouteTables"]

    assert len(verify[0]["Associations"]) == 1

    assert verify[0]["Associations"][0]["Main"] is False
    assert verify[0]["Associations"][0]["GatewayId"] == igw_id
    assert verify[0]["Associations"][0]["RouteTableAssociationId"] == assoc_id
    assert "SubnetId" not in verify[0]["Associations"][0]


@mock_aws
def test_associate_route_table_by_subnet():
    ec2 = boto3.client("ec2", region_name="us-west-1")
    vpc_id = ec2.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    route_table_id = ec2.create_route_table(VpcId=vpc_id)["RouteTable"]["RouteTableId"]
    subnet_id = ec2.create_subnet(VpcId=vpc_id, CidrBlock="10.0.0.0/24")["Subnet"][
        "SubnetId"
    ]
    assoc_id = ec2.associate_route_table(
        RouteTableId=route_table_id, SubnetId=subnet_id
    )["AssociationId"]
    verify = ec2.describe_route_tables(
        Filters=[
            {"Name": "association.route-table-association-id", "Values": [assoc_id]}
        ]
    )["RouteTables"]

    assert len(verify[0]["Associations"]) == 1

    assert verify[0]["Associations"][0]["Main"] is False
    assert verify[0]["Associations"][0]["SubnetId"] == subnet_id
    assert verify[0]["Associations"][0]["RouteTableAssociationId"] == assoc_id
    assert "GatewayId" not in verify[0]["Associations"][0]


def setup_vpc():
    ec2_resource = boto3.resource("ec2", region_name="eu-central-1")
    ec2_client = boto3.client("ec2", region_name="eu-central-1")

    vpc = ec2_resource.create_vpc(CidrBlock="10.0.0.0/16")
    ec2_resource.create_subnet(
        VpcId=vpc.id, CidrBlock="10.0.0.0/24", AvailabilityZone="us-west-2a"
    )

    route_table = ec2_resource.create_route_table(VpcId=vpc.id)

    return ec2_resource, ec2_client, route_table, vpc



import os
from functools import wraps
from uuid import uuid4

import boto3

from moto import mock_aws


def lakeformation_aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    Note that LakeFormation is not enabled by default - visit the AWS Console to permit access to the user who executes these tests.
    """

    @wraps(func)
    def pagination_wrapper():
        bucket_name = str(uuid4())

        allow_aws_request = (
            os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
        )

        if allow_aws_request:
            resp = create_glue_infra_and_test(bucket_name)
        else:
            with mock_aws():
                resp = create_glue_infra_and_test(bucket_name)
        return resp

    def create_glue_infra_and_test(bucket_name):
        glue = boto3.client("glue", region_name="eu-west-2")
        lf = boto3.client("lakeformation", region_name="eu-west-2")
        s3 = boto3.client("s3", region_name="us-east-1")

        s3.create_bucket(Bucket=bucket_name)
        s3.put_bucket_tagging(
            Bucket=bucket_name,
            Tagging={"TagSet": [{"Key": "environment", "Value": "moto_tests"}]},
        )
        lf.register_resource(
            ResourceArn=f"arn:aws:s3:::{bucket_name}", UseServiceLinkedRole=True
        )

        db_name = str(uuid4())[0:6]
        table_name = str(uuid4())[0:6]
        column_name = str(uuid4())[0:6]
        glue.create_database(
            DatabaseInput={"Name": db_name}, Tags={"environment": "moto_tests"}
        )
        glue.create_table(
            DatabaseName=db_name,
            TableInput={
                "Name": table_name,
                "StorageDescriptor": {
                    "Columns": [{"Name": column_name, "Type": "string"}]
                },
            },
        )

        try:
            resp = func(bucket_name, db_name, table_name, column_name)
        finally:
            ### CLEANUP ###

            glue.delete_table(DatabaseName=db_name, Name=table_name)
            glue.delete_database(Name=db_name)

            lf.deregister_resource(ResourceArn=f"arn:aws:s3:::{bucket_name}")

            versions = s3.list_object_versions(Bucket=bucket_name).get("Versions", [])
            for key in versions:
                s3.delete_object(
                    Bucket=bucket_name, Key=key["Key"], VersionId=key.get("VersionId")
                )
            delete_markers = s3.list_object_versions(Bucket=bucket_name).get(
                "DeleteMarkers", []
            )
            for key in delete_markers:
                s3.delete_object(
                    Bucket=bucket_name, Key=key["Key"], VersionId=key.get("VersionId")
                )
            s3.delete_bucket(Bucket=bucket_name)

        return resp

    return pagination_wrapper



"""Handles incoming qldb requests, invokes methods, returns responses."""

import json
from datetime import datetime
from typing import Dict, Optional, Union

from moto.core.responses import BaseResponse

from .models import QLDBBackend, qldb_backends


class QLDBResponse(BaseResponse):
    """Handler for QLDB requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="qldb")

    @property
    def qldb_backend(self) -> QLDBBackend:
        """Return backend instance specific for this region."""
        return qldb_backends[self.current_account][self.region]

    def _format_encryption_description(
        self, encryption_description: Optional[Dict[str, Union[str, datetime]]]
    ) -> Dict[str, Optional[Union[str, datetime]]]:
        return {
            "KmsKeyArn": (encryption_description or {}).get("kms_key_arn"),
            "EncryptionStatus": (encryption_description or {}).get("encryption_status"),
            "InaccessibleKmsKeyDateTime": (encryption_description or {}).get(
                "inaccessible_kms_key_date_time"
            ),
        }

    def describe_ledger(self) -> str:
        name = self._get_param("name")
        (
            name,
            arn,
            state,
            creation_date_time,
            permissions_mode,
            deletion_protection,
            encryption_description,
        ) = self.qldb_backend.describe_ledger(
            name=name,
        )
        creation_date_time_string = (
            creation_date_time.strftime("%d/%m/%Y, %H:%M:%S")
            if creation_date_time is not None
            else ""
        )
        return json.dumps(
            dict(
                Name=name,
                Arn=arn,
                State=state,
                CreationDateTime=creation_date_time_string,
                PermissionsMode=permissions_mode,
                DeletionProtection=deletion_protection,
                EncryptionDescription=self._format_encryption_description(
                    encryption_description
                ),
            )
        )

    def create_ledger(self) -> str:
        params = json.loads(self.body)
        name = params.get("Name")
        tags = params.get("Tags")
        permissions_mode = params.get("PermissionsMode")
        deletion_protection = params.get("DeletionProtection")
        kms_key = params.get("KmsKey")
        (
            name,
            arn,
            state,
            creation_date_time,
            permissions_mode,
            deletion_protection,
            kms_key_arn,
        ) = self.qldb_backend.create_ledger(
            name=name,
            tags=tags,
            permissions_mode=permissions_mode,
            deletion_protection=deletion_protection,
            kms_key=kms_key,
        )
        return json.dumps(
            dict(
                Name=name,
                Arn=arn,
                State=state,
                CreationDateTime=creation_date_time.strftime("%d/%m/%Y, %H:%M:%S"),
                PermissionsMode=permissions_mode,
                DeletionProtection=deletion_protection,
                KmsKeyArn=kms_key_arn,
            )
        )

    def delete_ledger(self) -> str:
        name = self._get_param("name")
        self.qldb_backend.delete_ledger(
            name=name,
        )
        return json.dumps(dict())

    def update_ledger(self) -> str:
        name = self._get_param("name")
        params = json.loads(self.body)
        deletion_protection = params.get("DeletionProtection")
        kms_key = params.get("KmsKey")
        (
            name,
            arn,
            state,
            creation_date_time,
            deletion_protection,
            encryption_description,
        ) = self.qldb_backend.update_ledger(
            name=name,
            deletion_protection=deletion_protection,
            kms_key=kms_key,
        )
        creation_date_time_string = (
            creation_date_time.strftime("%d/%m/%Y, %H:%M:%S")
            if creation_date_time is not None
            else ""
        )
        return json.dumps(
            dict(
                name=name,
                arn=arn,
                state=state,
                creationDateTime=creation_date_time_string,
                deletionProtection=deletion_protection,
                encryptionDescription=self._format_encryption_description(
                    encryption_description
                ),
            )
        )

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = params.get("ResourceArn")
        tags = params.get("Tags")
        self.qldb_backend.tag_resource(
            resource_arn=resource_arn,
            new_tags=tags,
        )
        return json.dumps(dict())

    def list_tags_for_resource(self) -> str:
        resource_arn = self._get_param("ResourceArn")
        tags = self.qldb_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(Tags=tags))



import json
from typing import TYPE_CHECKING, Any, Dict, List, TypedDict

if TYPE_CHECKING:
    from typing_extensions import Any, Dict, Required, Union


# NOTE: Typing is based on the following document https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-event-patterns.html
EventMessageType = TypedDict(
    "EventMessageType",
    {
        "version": str,
        "id": str,
        "detail-type": "Required[Union[str, List[str]]]",
        "source": "Required[Union[str, List[str]]]",
        "account": str,
        # support float type for internal use of moto.
        "time": "Union[str, float]",
        "replay-name": str,
        "region": str,
        "resources": List[str],
        "detail": "Required[Dict[str, Any]]",
        "ingestion-time": float,
    },
    total=False,
)

PAGINATION_MODEL = {
    "list_rules": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 50,
        "unique_attribute": "arn",
        "fail_on_invalid_token": False,
    },
    "list_rule_names_by_target": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 50,
        "unique_attribute": "arn",
        "fail_on_invalid_token": False,
    },
    "list_targets_by_rule": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 50,
        "unique_attribute": "Arn",
        "fail_on_invalid_token": False,
    },
}

_BASE_EVENT_MESSAGE: EventMessageType = {
    "version": "0",
    "id": "17793124-05d4-b198-2fde-7ededc63b103",
    "detail-type": "",
    "source": "",
    "account": "",
    "time": "",
    "region": "",
    "resources": [],
    "detail": {},
}


class EventTemplateParser:
    DEFAULT_EVENT_INPUT_TEMPLATE = '{"id": "<id>", "time": <time>, "version": "0", "detail-type": "<detail-type>", "source": "<source>", "region": "<region>", "resources": <resources>, "detail": <detail>}'
    DEFAULT_EVENT_INPUT_PATHS_MAP = {
        "account": "$.account",
        "detail": "$.detail",
        "detail-type": "$.detail-type",
        "source": "$.source",
        "id": "$.id",
        "region": "$.region",
        "resources": "$.resources",
        "time": "$.time",
    }

    @staticmethod
    def _stringify(result: Any) -> str:  # type: ignore[misc]
        if isinstance(result, dict):
            result = json.dumps(result)
        elif isinstance(result, list):
            result = json.dumps([EventTemplateParser._stringify(x) for x in result])
        elif isinstance(result, (int, float)):
            result = str(result)
        return result

    @staticmethod
    def parse(  # type: ignore[misc]
        input_template: str, input_paths_map: Dict[str, Any], event: EventMessageType
    ) -> Dict[str, Any]:
        from jsonpath_ng.ext import parse

        template_to_use = (
            input_template or EventTemplateParser.DEFAULT_EVENT_INPUT_TEMPLATE
        )
        input_paths_map = (
            input_paths_map or EventTemplateParser.DEFAULT_EVENT_INPUT_PATHS_MAP
        )
        for input_path in input_paths_map:
            input_expr = parse(input_paths_map[input_path])
            matches = input_expr.find(event)
            result = (
                EventTemplateParser._stringify(matches[0].value) if matches else None
            )
            if result:
                template_to_use = template_to_use.replace(f"<{input_path}>", result)

        default_inputs_map = {
            "aws.events.event.json": event,
            "aws.events.event.ingestion-time": event["ingestion-time"],
        }
        for input_path in default_inputs_map:
            result = EventTemplateParser._stringify(default_inputs_map[input_path])
            template_to_use = template_to_use.replace(f"<{input_path}>", result)

        return json.loads(template_to_use)



from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional

from moto.appmesh.dataclasses.shared import (
    Duration,
    Metadata,
    MissingField,
    Status,
    Timeout,
)
from moto.appmesh.utils.common import clean_dict


@dataclass
class RouteActionWeightedTarget:
    virtual_node: str
    weight: int
    port: Optional[int] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"port": self.port, "virtualNode": self.virtual_node, "weight": self.weight}
        )


@dataclass
class RouteAction:
    weighted_targets: List[RouteActionWeightedTarget]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"weightedTargets": [target.to_dict() for target in self.weighted_targets]}
        )


@dataclass
class Range:
    start: int
    end: int
    to_dict = asdict


@dataclass
class Match:
    exact: Optional[str]
    prefix: Optional[str]
    range: Optional[Range]
    regex: Optional[str]
    suffix: Optional[str]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "exact": self.exact,
                "prefix": self.prefix,
                "range": (self.range or MissingField()).to_dict(),
                "regex": self.regex,
                "suffix": self.suffix,
            }
        )


@dataclass
class GrpcMetadatum:
    invert: Optional[bool]
    match: Optional[Match]
    name: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "invert": self.invert,
                "match": (self.match or MissingField()).to_dict(),
                "name": self.name,
            }
        )


# same object, just different name
HttpRouteMatchHeader = GrpcMetadatum


@dataclass
class RouteMatchPath:
    exact: str
    regex: str
    to_dict = asdict


@dataclass
class QueryParameterMatch:
    exact: str
    to_dict = asdict


@dataclass
class RouteMatchQueryParameter:
    name: str
    match: Optional[QueryParameterMatch] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"match": (self.match or MissingField()).to_dict(), "name": self.name}
        )


@dataclass
class HttpRouteMatch:
    headers: Optional[List[HttpRouteMatchHeader]] = field(default=None)
    method: Optional[str] = field(default=None)
    path: Optional[RouteMatchPath] = field(default=None)
    port: Optional[int] = field(default=None)
    prefix: Optional[str] = field(default=None)
    query_parameters: Optional[List[RouteMatchQueryParameter]] = field(default=None)
    scheme: Optional[str] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "headers": [header.to_dict() for header in self.headers or []],
                "method": self.method,
                "path": (self.path or MissingField()).to_dict(),
                "port": self.port,
                "prefix": self.prefix,
                "queryParameters": [
                    param.to_dict() for param in self.query_parameters or []
                ],
                "scheme": self.scheme,
            }
        )


@dataclass
class HttpRouteRetryPolicy:
    max_retries: int
    http_retry_events: Optional[List[str]]
    per_retry_timeout: Duration
    tcp_retry_events: Optional[List[str]]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "httpRetryEvents": self.http_retry_events or [],
                "maxRetries": self.max_retries,
                "perRetryTimeout": self.per_retry_timeout.to_dict(),
                "tcpRetryEvents": self.tcp_retry_events or [],
            }
        )


@dataclass
class GrpcRouteMatch:
    metadata: Optional[List[GrpcMetadatum]]
    method_name: Optional[str]
    port: Optional[int]
    service_name: Optional[str]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "metadata": [meta.to_dict() for meta in self.metadata or []],
                "methodName": self.method_name,
                "port": self.port,
                "serviceName": self.service_name,
            }
        )


@dataclass
class GrcpRouteRetryPolicy:
    max_retries: int
    per_retry_timeout: Duration
    grpc_retry_events: Optional[List[str]]
    http_retry_events: Optional[List[str]]
    tcp_retry_events: Optional[List[str]]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "grpcRetryEvents": self.grpc_retry_events or [],
                "httpRetryEvents": self.http_retry_events or [],
                "maxRetries": self.max_retries,
                "perRetryTimeout": self.per_retry_timeout.to_dict(),
                "tcpRetryEvents": self.tcp_retry_events or [],
            }
        )


@dataclass
class GrpcRoute:
    action: RouteAction
    match: GrpcRouteMatch
    retry_policy: Optional[GrcpRouteRetryPolicy]
    timeout: Optional[Timeout]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "action": self.action.to_dict(),
                "match": self.match.to_dict(),
                "retryPolicy": (self.retry_policy or MissingField()).to_dict(),
                "timeout": (self.timeout or MissingField()).to_dict(),
            }
        )


@dataclass
class HttpRoute:
    action: RouteAction
    match: HttpRouteMatch
    retry_policy: Optional[HttpRouteRetryPolicy] = field(default=None)
    timeout: Optional[Timeout] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "action": self.action.to_dict(),
                "match": self.match.to_dict(),
                "retryPolicy": (self.retry_policy or MissingField()).to_dict(),
                "timeout": (self.timeout or MissingField()).to_dict(),
            }
        )


@dataclass
class TCPRouteMatch:
    port: int
    to_dict = asdict


@dataclass
class TCPRoute:
    action: RouteAction
    match: Optional[TCPRouteMatch]
    timeout: Optional[Timeout]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "action": self.action.to_dict(),
                "match": (self.match or MissingField()).to_dict(),
                "timeout": (self.timeout or MissingField()).to_dict(),
            }
        )


@dataclass
class RouteSpec:
    priority: Optional[int]
    grpc_route: Optional[GrpcRoute] = field(default=None)
    http_route: Optional[HttpRoute] = field(default=None)
    http2_route: Optional[HttpRoute] = field(default=None)
    tcp_route: Optional[TCPRoute] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        spec = {
            "grpcRoute": (self.grpc_route or MissingField()).to_dict(),
            "httpRoute": (self.http_route or MissingField()).to_dict(),
            "http2Route": (self.http2_route or MissingField()).to_dict(),
            "priority": self.priority,
            "tcpRoute": (self.tcp_route or MissingField()).to_dict(),
        }
        return clean_dict(spec)


@dataclass
class RouteMetadata(Metadata):
    mesh_name: str = field(default="")
    route_name: str = field(default="")
    virtual_router_name: str = field(default="")

    def __post_init__(self) -> None:
        if self.mesh_name == "":
            raise TypeError("__init__ missing 1 required argument: 'mesh_name'")
        if self.mesh_owner == "":
            raise TypeError("__init__ missing 1 required argument: 'route_name'")
        if self.virtual_router_name == "":
            raise TypeError(
                "__init__ missing 1 required argument: 'virtual_router_name'"
            )

    def formatted_for_list_api(self) -> Dict[str, Any]:  # type: ignore
        return {
            "arn": self.arn,
            "createdAt": self.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "lastUpdatedAt": self.last_updated_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "meshName": self.mesh_name,
            "meshOwner": self.mesh_owner,
            "resourceOwner": self.resource_owner,
            "routeName": self.route_name,
            "version": self.version,
            "virtualRouterName": self.virtual_router_name,
        }

    def formatted_for_crud_apis(self) -> Dict[str, Any]:  # type: ignore
        return {
            "arn": self.arn,
            "createdAt": self.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "lastUpdatedAt": self.last_updated_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "meshOwner": self.mesh_owner,
            "resourceOwner": self.resource_owner,
            "uid": self.uid,
            "version": self.version,
        }


@dataclass
class Route:
    mesh_name: str
    mesh_owner: str
    metadata: RouteMetadata
    route_name: str
    spec: RouteSpec
    virtual_router_name: str
    status: Status = field(default_factory=lambda: {"status": "ACTIVE"})
    tags: List[Dict[str, str]] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "meshName": self.mesh_name,
                "metadata": self.metadata.formatted_for_crud_apis(),
                "routeName": self.route_name,
                "spec": self.spec.to_dict(),
                "status": self.status,
                "tags": self.tags,
                "virtualRouterName": self.virtual_router_name,
            }
        )
